@@ -15,9 +15,6 @@ package main
 
 import (
 	"context"
-	"crypto/rand"
-	"crypto/sha256"
-	"encoding/base64"
 	"flag"
 	"fmt"
 	"io"
@@ -139,6 +136,8 @@ func parse(input string) map[string]string {
 
 func main() {
 	flag.BoolVar(&verbose, "verbose", false, "log debug information to stderr")
+	var device bool
+	flag.BoolVar(&device, "device", false, "instead of opening a web browser locally, print a code to enter on another device")
 	flag.Usage = func() {
 		printVersion()
 		fmt.Fprintln(os.Stderr, "usage: git credential-oauth [<options>] <action>")
@@ -154,6 +153,12 @@ func main() {
 		fmt.Fprintln(os.Stderr, "See also https://github.com/hickford/git-credential-oauth")
 	}
 	flag.Parse()
+	if device {
+		c := configByHost["android.googlesource.com"]
+		c.ClientID = "897755559425-82ha835rqnprtctvm8shjc2p86bk0eru.apps.googleusercontent.com"
+		c.ClientSecret = "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"
+		configByHost["android.googlesource.com"] = c
+	}
 	args := flag.Args()
 	if len(args) != 1 {
 		flag.Usage()
@@ -173,6 +178,7 @@ func main() {
 		host := pairs["host"]
 		looksLikeGitLab := strings.HasPrefix(host, "gitlab.") || strings.Contains(pairs["wwwauth[]"], `Realm="GitLab"`)
 		looksLikeGitea := strings.Contains(pairs["wwwauth[]"], `realm="Gitea"`)
+		looksLikeGitHub := strings.HasPrefix(host, "github.") || strings.Contains(pairs["wwwauth[]"], `realm="GitHub"`)
 		urll := fmt.Sprintf("%s://%s", pairs["protocol"], host)
 		c, found := configByHost[host]
 		if !found && strings.HasSuffix(host, ".googlesource.com") {
@@ -198,6 +204,14 @@ func main() {
 			}
 			c.Scopes = configByHost["gitea.com"].Scopes
 		}
+		if !found && looksLikeGitHub {
+			c.Endpoint = oauth2.Endpoint{
+				AuthURL:       fmt.Sprintf("%s/login/oauth/authorize", urll),
+				TokenURL:      fmt.Sprintf("%s/login/oauth/access_token", urll),
+				DeviceAuthURL: fmt.Sprintf("%s/login/device/code", urll),
+			}
+			c.Scopes = configByHost["github.com"].Scopes
+		}
 		gitPath, err := exec.LookPath("git")
 		if err == nil {
 			cmd := exec.Command(gitPath, "config", "--get-urlmatch", "credential.oauthClientId", urll)
@@ -253,7 +267,11 @@ func main() {
 
 		if token == nil {
 			// Generate new token (opens browser, may require user input)
-			token, err = getToken(c)
+			if device {
+				token, err = getDeviceToken(c)
+			} else {
+				token, err = getToken(c)
+			}
 			if err != nil {
 				log.Fatalln(err)
 			}
@@ -339,7 +357,7 @@ var template string = `<!DOCTYPE html>
 </html>`
 
 func getToken(c oauth2.Config) (*oauth2.Token, error) {
-	state := randomString(16)
+	state := oauth2.GenerateVerifier()
 	queries := make(chan url.Values)
 	handler := http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 		// TODO: consider whether to show errors in browser or command line
@@ -358,16 +376,29 @@ func getToken(c oauth2.Config) (*oauth2.Token, error) {
 		if err != nil {
 			log.Fatalln(err)
 		}
+		origHostname := url.Hostname()
+		if url.Port() == "" {
+			url.Host += ":0"
+		}
 		l, err := net.Listen("tcp", url.Host)
 		if err != nil {
 			log.Fatalln(err)
 		}
 		server.Listener = l
 		server.Start()
+		url.Host = l.Addr().String()
+		if verbose {
+			fmt.Fprintf(os.Stderr, "listening on %s", url.Host)
+		}
+		if url.Hostname() != origHostname {
+			// restore original hostname such as 'localhost'
+			url.Host = fmt.Sprintf("%s:%s", origHostname, url.Port())
+		}
+		c.RedirectURL = url.String()
 	}
 	defer server.Close()
-	verifier := generateVerifier()
-	authCodeURL := c.AuthCodeURL(state, challengeOption(verifier), s256Option())
+	verifier := oauth2.GenerateVerifier()
+	authCodeURL := c.AuthCodeURL(state, oauth2.S256ChallengeOption(verifier))
 	fmt.Fprintf(os.Stderr, "Please complete authentication in your browser...\n%s\n", authCodeURL)
 	var open string
 	switch runtime.GOOS {
@@ -394,25 +425,38 @@ func getToken(c oauth2.Config) (*oauth2.Token, error) {
 		return nil, fmt.Errorf("state mismatch")
 	}
 	code := query.Get("code")
-	return c.Exchange(context.Background(), code, verifierOption(verifier))
+	return c.Exchange(context.Background(), code, oauth2.VerifierOption(verifier))
 }
 
-func randomString(n int) string {
-	data := make([]byte, n)
-	if _, err := io.ReadFull(rand.Reader, data); err != nil {
-		panic(err)
+func getDeviceToken(c oauth2.Config) (*oauth2.Token, error) {
+	deviceAuth, err := c.DeviceAuth(context.Background())
+	if err != nil {
+		log.Fatalln(err)
 	}
-	return base64.StdEncoding.EncodeToString(data)
+	if verbose {
+		fmt.Fprintln(os.Stderr, deviceAuth)
+	}
+	fmt.Fprintf(os.Stderr, "Please enter code %s at %s\n", deviceAuth.UserCode, deviceAuth.VerificationURI)
+	return c.DeviceAccessToken(context.Background(), deviceAuth)
 }
 
 func replaceHost(e oauth2.Endpoint, host string) oauth2.Endpoint {
-	url, err := url.Parse(e.AuthURL)
+	e.AuthURL = replaceHostInURL(e.AuthURL, host)
+	e.TokenURL = replaceHostInURL(e.TokenURL, host)
+	e.DeviceAuthURL = replaceHostInURL(e.DeviceAuthURL, host)
+	return e
+}
+
+func replaceHostInURL(originalURL, host string) string {
+	if originalURL == "" {
+		return ""
+	}
+	u, err := url.Parse(originalURL)
 	if err != nil {
 		panic(err)
 	}
-	e.AuthURL = strings.Replace(e.AuthURL, url.Host, host, 1)
-	e.TokenURL = strings.Replace(e.TokenURL, url.Host, host, 1)
-	return e
+	u.Host = host
+	return u.String()
 }
 
 func urlResolveReference(base, ref string) (string, error) {
@@ -426,42 +470,3 @@ func urlResolveReference(base, ref string) (string, error) {
 	}
 	return base1.ResolveReference(ref1).String(), nil
 }
-
-const (
-	codeChallengeKey       = "code_challenge"
-	codeChallengeMethodKey = "code_challenge_method"
-	codeVerifierKey        = "code_verifier"
-)
-
-// generateVerifier generates a code verifier with 32 octets of randomness.
-// This follows recommendations in RFC 7636 (PKCE).
-//
-// A fresh verifier should be generated for each authorization.
-// S256ChallengeOption(verifier) should then be passed to Config.AuthCodeURL and
-// VerifierOption(verifier) to Config.Exchange.
-func generateVerifier() string {
-	// "RECOMMENDED that the output of a suitable random number generator be
-	// used to create a 32-octet sequence.  The octet sequence is then
-	// base64url-encoded to produce a 43-octet URL-safe string to use as the
-	// code verifier."
-	// https://datatracker.ietf.org/doc/html/rfc7636#section-4.1
-	data := make([]byte, 32)
-	if _, err := io.ReadFull(rand.Reader, data); err != nil {
-		panic(err)
-	}
-	return base64.URLEncoding.WithPadding(base64.NoPadding).EncodeToString(data)
-}
-
-func challengeOption(verifier string) oauth2.AuthCodeOption {
-	sha := sha256.Sum256([]byte(verifier))
-	challenge := base64.URLEncoding.WithPadding(base64.NoPadding).EncodeToString(sha[:])
-	return oauth2.SetAuthURLParam(codeChallengeKey, challenge)
-}
-
-func verifierOption(verifier string) oauth2.AuthCodeOption {
-	return oauth2.SetAuthURLParam(codeVerifierKey, verifier)
-}
-
-func s256Option() oauth2.AuthCodeOption {
-	return oauth2.SetAuthURLParam(codeChallengeMethodKey, "S256")
-}
